/* eslint-disable @typescript-eslint/no-explicit-any */
import { AccessSchema, Permit, Schema } from '../permit';

export class ProcedureError extends Error {
  path: string;

  constructor(path: string, message: string, error: Error) {
    super(message);
    Object.setPrototypeOf(this, ProcedureError.prototype);

    this.path = path;
    this.stack = error.stack;
    this.cause = error.cause;
  }
}

export type onError = (error: ProcedureError) => void;
export type onCall<T> = (
  path: string,
  context: ContextValue<T, any>,
  input: any[]
) => void;

/* eslint-disable @typescript-eslint/no-explicit-any */
export type ContextRoot = { scopes?: string[]; protected?: boolean };
export type Context<T> = T & ContextRoot;
export type ContextValue<T, V> = T & { value: V } & ContextRoot;

export type GuardFunction<S extends Schema> = (permit: Permit<S>) => boolean;

export type Pipe<T, I extends any[], V> = {
  fn: (...input: I) => V;
  pipe: <Output>(
    fn: (context: ContextValue<T, V>, ...input: I) => Output
  ) => Pipe<T, I, Output>;
  protect: <S extends Schema>(
    schema: AccessSchema<S>,
    guard: GuardFunction<S>
  ) => Pipe<T, I, V>;
  permit: (scopes: string[]) => Pipe<T, I, V>;
};

export type InterfaceConfig<T> = {
  immidiate?: boolean;
  onError?: onError;
  onCall?: onCall<T>;
};

export type Interface<T> = {
  procedure: <I extends any[], V = void>(
    key: string,
    fn: (context: Context<T>, ...input: I) => V
  ) => Pipe<T, I, V>;
  protect: <S extends Schema>(
    schema: AccessSchema<S>,
    guard: GuardFunction<S>
  ) => Omit<Interface<T>, 'protect'>;
};
